'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var _getIterator = _interopDefault(require('babel-runtime/core-js/get-iterator'));
var _classCallCheck = _interopDefault(require('babel-runtime/helpers/classCallCheck'));
var _createClass = _interopDefault(require('babel-runtime/helpers/createClass'));
var _regeneratorRuntime = _interopDefault(require('babel-runtime/regenerator'));
var _defineProperty = _interopDefault(require('babel-runtime/helpers/defineProperty'));
var _extends = _interopDefault(require('babel-runtime/helpers/extends'));
var _Object$assign = _interopDefault(require('babel-runtime/core-js/object/assign'));
var _Object$keys = _interopDefault(require('babel-runtime/core-js/object/keys'));
var _typeof = _interopDefault(require('babel-runtime/helpers/typeof'));

/**
 * Js Immutable
 *
 * Copyright © 2015-2016 Robus Gauli, LLC. All rights reserved.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE.txt file in the root directory of this source tree.
 */

/**
 * Accepts the selector object and operation intended
 *
 * ### Example (es module)
 * ```js
 * import reduce from 'js-immutable';
 * const deepObject = {
 *  person: {
 *    name: 'robus gauli',
 *    address: {
 *      permanent: 'Nepal',
 *      temporary: 'China'
 *    }
 *  }
 * }
 *
 * const selector = {
 *  person: {
 *    address: {
 *      permanent: '#'
 *    }
 *  }
 * }
 *
 * const reducer = reduce(selector)
 * const result = reducer(deepObject)
 *                  .set('Holy land')
 *                  .apply()
 * // => {
 *    person: {
 *      name: ...,
 *      address: {
 *        permanent: 'Holy land',
 *        temporary: 'China'
 *      }
 *    }
 *  }
 * ```
 */

function reduce(selector) {
  if (selector === null || !((typeof selector === 'undefined' ? 'undefined' : _typeof(selector)) === 'object') || Array.isArray(selector)) {
    throw new Error('Invalid selector');
  }

  if (!_Object$keys(selector).length) {
    throw new Error('Invalid selector. No keys to traverse.');
  }
  return function selectorWrapper(_originalObject) {
    var _marked = /*#__PURE__*/_regeneratorRuntime.mark(toUpdateCoro);

    function coroutine(coroutineFunction) {
      return function wrapper() {
        // launch the coroutine
        var coro = coroutineFunction.apply(undefined, arguments);
        // drive it to the first yield
        coro.next();
        return coro;
      };
    }

    function set(oldValue, newValue) {
      if (newValue === null || newValue === undefined) {
        return oldValue;
      }
      return newValue;
    }

    function extend(oldArray, newArray) {
      if (!Array.isArray(newArray) || !Array.isArray(oldArray)) {
        return oldArray;
      }
      if (newArray.length) {
        return oldArray.concat(newArray);
      }
      return oldArray;
    }

    function append(oldList, value) {
      // if value is null or undefined return the same original list
      if (value === undefined || value === null || !Array.isArray(oldList)) {
        return oldList;
      }
      return oldList.concat(value);
    }

    function merge(oldObject, newObject) {
      if (newObject === null || newObject === undefined || (typeof newObject === 'undefined' ? 'undefined' : _typeof(newObject)) !== 'object') {
        return oldObject;
      }
      if ((typeof newObject === 'undefined' ? 'undefined' : _typeof(newObject)) === 'object' && !_Object$keys(newObject).length) {
        return oldObject;
      }
      return _Object$assign({}, oldObject, newObject);
    }

    // delete being the builtin operator
    function deleteOp(oldObject, key) {
      if (key === null || key === undefined) {
        return oldObject;
      }
      if (Array.isArray(oldObject)) {
        return oldObject.filter(function (val, index) {
          return key !== index;
        });
      }
      if ((typeof oldObject === 'undefined' ? 'undefined' : _typeof(oldObject)) === 'object' && oldObject !== null) {
        return _Object$keys(oldObject).reduce(function (acc, k) {
          return k === key ? _extends({}, acc) : _extends({}, acc, _defineProperty({}, k, oldObject[k]));
        }, {});
      }
      return oldObject;
    }

    var operationalMapper = {
      set: set,
      append: append,
      merge: merge,
      extend: extend,
      delete: deleteOp
    };

    var toUpdate = {};
    var targetPointer = '#';

    function toUpdateCoro() {
      var _ref, updatePayload, pointer;

      return _regeneratorRuntime.wrap(function toUpdateCoro$(_context) {
        while (1) {
          switch (_context.prev = _context.next) {
            case 0:
              

              _context.next = 3;
              return;

            case 3:
              _ref = _context.sent;
              updatePayload = _ref.updatePayload;
              pointer = _ref.pointer;


              if (pointer === '#') {
                toUpdate = _extends({}, toUpdate, {
                  default: updatePayload
                });
              } else if (pointer.startsWith('#') && pointer.length > 1) {
                toUpdate = _extends({}, toUpdate, _defineProperty({}, pointer, updatePayload));
              }
              _context.next = 0;
              break;

            case 9:
            case 'end':
              return _context.stop();
          }
        }
      }, _marked, this);
    }

    // drive the coroutine to the first yield
    var updateCoroutine = coroutine(toUpdateCoro)();

    return new (function () {
      function _class() {
        _classCallCheck(this, _class);

        this.set = this.operationFactory('set');
        this.merge = this.operationFactory('merge');
        this.delete = this.operationFactory('delete');
        this.append = this.operationFactory('append');
        this.extend = this.operationFactory('extend');
      }

      _createClass(_class, [{
        key: 'of',
        value: function of(target) {
          if (!target) {
            return this;
          }
          if (typeof target === 'string' && target.startsWith('#') && target.length > 1) {
            targetPointer = target;
          }
          return this;
        }
      }, {
        key: 'operationFactory',
        value: function operationFactory(operationName) {
          return function wrapper(value) {
            if (value !== undefined || value !== null) {
              updateCoroutine.next({
                updatePayload: {
                  value: value,
                  operation: operationName
                },
                pointer: targetPointer
              });
            }
            return this;
          }.bind(this);
        }
      }, {
        key: 'selectTransform',
        value: function selectTransform(originalObject, selectorObject) {
          if (typeof selectorObject === 'string' && selectorObject.startsWith('#')) {
            // we reached to the target
            var updateKeys = _Object$keys(toUpdate);

            if (!updateKeys.length) {
              throw new Error('Should at least has a one operation intended before applying changes');
            }

            if (selectorObject === '#') {
              if (updateKeys.indexOf('default') !== -1) {
                var updater = toUpdate.default;
                return operationalMapper[updater.operation](originalObject, updater.value);
              }
              return originalObject;
            } else if (updateKeys.indexOf(selectorObject) !== -1) {
              var _updater = toUpdate[selectorObject];
              return operationalMapper[_updater.operation](originalObject, _updater.value);
            }
            return originalObject;
          }

          var clonedObject = _Object$assign({}, originalObject);

          var _iteratorNormalCompletion = true;
          var _didIteratorError = false;
          var _iteratorError = undefined;

          try {
            for (var _iterator = _getIterator(_Object$keys(selectorObject)), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
              var key = _step.value;

              var keyedObject = clonedObject[key];
              if (keyedObject === undefined || keyedObject === null) {
                // return back the original state and the copied
                clonedObject = originalObject;
                break;
              }
              var filteredObject = selectorObject[key];
              clonedObject[key] = this.selectTransform(keyedObject, filteredObject);
            }
          } catch (err) {
            _didIteratorError = true;
            _iteratorError = err;
          } finally {
            try {
              if (!_iteratorNormalCompletion && _iterator.return) {
                _iterator.return();
              }
            } finally {
              if (_didIteratorError) {
                throw _iteratorError;
              }
            }
          }

          return clonedObject;
        }
      }, {
        key: 'apply',
        value: function apply() {
          return this.selectTransform(_originalObject, selector);
        }
      }]);

      return _class;
    }())();
  };
}

module.exports = reduce;
//# sourceMappingURL=index.js.map
